/* tacacs.c v.0.5 Plugin for pppd. Implement TACACS+ protocol.
 * Copyright (C)-2000/2002 Jean-Louis Noel (jln@stben.be)
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <syslog.h>
#include <pppd.h>
#include <patchlevel.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <tacplus.h>
#include <libtac.h>
#include <ipcp.h>

bool   use_tacacs  = 0;
bool   use_authorize  = 0;
bool   use_account = 0;
int    tac_servers = 0;
int    tac_encryption = 1;
int    task_id;
u_long tac_server[3];
char   *tac_secret;
char   buf[256];
char   tty[256];
char   tac_sname[MAXNAMELEN];		/* ip name TACACS+ server */
char   tac_coding[MAXSECRETLEN];	/* key used for crypt TACACS+ paquet */
// static char   pppd_version[] = VERSION;	/*  */

static option_t tacacs_options[] =
    {
	{"tacacs", o_bool, &use_tacacs,
	    "Use TACACS+ functions", 1},
	{"tacacs-accounting", o_bool, &use_account,
	    "send TACACS+ accounting packets", 1},
	{"tacacs-authorization", o_bool, &use_authorize,
	    "Authorize user for use PPP", 1},
	{"tacacs-server", o_string, tac_sname,
	    "TACACS+ server IP address",OPT_STATIC,NULL,MAXNAMELEN},
	{"tacacs-secret", o_string, tac_coding,
	    "key used to encrypt TACACS+ packets",OPT_STATIC,NULL,MAXSECRETLEN},
	{NULL}
    };

static unsigned long
getservername(char *serv)
    {
    struct in_addr addr;
    struct hostent *h;

    if(inet_aton(serv, &addr) == 0)
	{
	if((h=gethostbyname(serv)) == NULL)
	    syslog(LOG_ERR,"TACACS+ gethostbyname failed: %s",hstrerror(h_errno));
	else
	    {
	    bcopy(h->h_addr, (char *)&addr, sizeof(struct in_addr));
	    return(addr.s_addr);
	    }
        }
    else return(addr.s_addr);

    return(-1);
    }

static int
tacacs_auth(t_user, t_passwd, t_msgp, t_paddrs, t_popts)

char *t_user;
char *t_passwd;
char **t_msgp;
struct wordlist **t_paddrs;
struct wordlist **t_popts;
    {
    int  tac_fd;
    char *msg;
    char *ptr;
    struct areply   arep;
    struct tac_attrib *parep;
    struct wordlist *ap;
    struct tac_attrib *attr;

    if (!use_tacacs) return -1;

    *t_msgp = "User/Password Incorrect\0";
    *t_popts = NULL;
    bzero(&arep,sizeof(struct areply));

    /* start authentication */
    
    tac_fd=tac_connect(&tac_server[0], tac_servers);
    if(tac_fd < 0)
	{
	syslog(LOG_ERR,"TACACS+ server \"%s\" failed", tac_sname);
	return -1;
	}

    if(tac_authen_pap_send(tac_fd, t_user, t_passwd, tty) < 0) return -1;

    msg = tac_authen_pap_read(tac_fd);
    if(msg != NULL) return 0;

    close(tac_fd);

    /* Name/Password valid */

    if (use_authorize)
	{
	tac_fd=tac_connect(&tac_server[0], tac_servers);
    	if(tac_fd < 0)
	    {
	    syslog(LOG_ERR,"TACACS+ server \"%s\" failed", tac_sname);
	    return -1;
	    }
	attr = NULL;
                
	tac_add_attrib(&attr, "service", "ppp");
	tac_add_attrib(&attr, "protocol", "ip");

	tac_author_send(tac_fd, t_user, tty, attr);

	tac_author_read(tac_fd, &arep);

	if (arep.status != AUTHOR_STATUS_PASS_ADD &&
	    arep.status != AUTHOR_STATUS_PASS_REPL)
	    {
	    *t_msgp = "Service not authorized\0";
    	    return 0;
    	    }
                                                                                                                        
	tac_free_attrib(&attr);
	close(tac_fd);
	}

    
    *t_msgp = "Login succeded\0";
    parep = arep.attr;
    syslog(LOG_INFO,"TACACS+ login succeded for %s", t_user);
    ap = (struct wordlist *) malloc(sizeof(struct wordlist));

    if (!ap) novm("TACACS+ address");

    msg = NULL;
    if (parep) msg = strchr(arep.attr->attr,'=');

    if (msg) msg[0] = '\0';
    msg++;
    ap->word = strdup("*\0");
    if(parep)
	{
	if (!strcmp(arep.attr->attr,"addr"))
	    ap->word = strdup(msg);
	}
    if (!ap->word) novm("TACACS+ options");
    ptr = strchr(ap->word,25);
    if (ptr) ptr[0] = '\0';
    ap->next = NULL;
    if (t_paddrs) *t_paddrs = ap;
    return 1;
    }


static void
accounting_start(void)
    {
    int  tac_fd;
    char *phone;
    char *msg;
    struct tac_attrib   *attr;
    struct in_addr      peer_addr;
    
    if(use_tacacs & use_account)
	{
	phone = getenv("CALLER_ID");
	if (!phone) phone = "Unknow";
	tac_fd=tac_connect(&tac_server[0], tac_servers);
	if(tac_fd < 0)
	    {
	    syslog(LOG_ERR,"TACACS+ server %s failed", tac_sname);
	    return;
	    }

	/* start accounting */
	attr = NULL;

	sprintf(buf, "%lu", time(0));
	tac_add_attrib(&attr, "start_time", buf);
	sprintf(buf, "%hu", task_id);
	tac_add_attrib(&attr, "task_id", buf);
	tac_add_attrib(&attr, "phone_#", phone);
	tac_add_attrib(&attr, "service", "ppp");
	tac_add_attrib(&attr, "protocol", "ip");
	peer_addr.s_addr = ipcp_hisoptions[0].hisaddr;
	sprintf(buf, "%s", inet_ntoa(peer_addr));

	tac_account_send(tac_fd, TAC_PLUS_ACCT_FLAG_START, peer_authname, tty, buf, attr);

	msg = tac_account_read(tac_fd);
	if(msg != NULL)
	    syslog(LOG_ERR,"TACACS+ start accounting failed: %s", msg);

	tac_free_attrib(&attr);
	close(tac_fd); 
	}
    }

static void
accounting_stop(void)
    {
    int  tac_fd;
    char *msg;
    struct tac_attrib *attr;
    struct in_addr      peer_addr;

    if(use_tacacs & use_account)
	{
	tac_fd=tac_connect(&tac_server[0], tac_servers);
	if(tac_fd < 0)
	    {
	    syslog(LOG_ERR,"TACACS+ server %s failed", tac_sname);
	    return;
	    }

	/* stop accounting */
	attr = NULL;

	sprintf(buf, "%lu", time(0));
	tac_add_attrib(&attr, "stop_time", buf);
	sprintf(buf, "%hu", task_id);
	tac_add_attrib(&attr, "task_id", buf);
	if (link_stats_valid)
	    {
	    sprintf(buf, "%d", link_stats.bytes_out);
	    tac_add_attrib(&attr, "bytes_out", buf);
	    sprintf(buf, "%d", link_stats.bytes_in);
	    tac_add_attrib(&attr, "bytes_in", buf);
	    sprintf(buf, "%d", link_connect_time);
	    tac_add_attrib(&attr, "elapsed_time", buf);
	    peer_addr.s_addr = ipcp_hisoptions[0].hisaddr;
	    sprintf(buf, "%s", inet_ntoa(peer_addr));
	    }

	tac_account_send(tac_fd, TAC_PLUS_ACCT_FLAG_STOP, peer_authname, tty, buf, attr);

	msg = tac_account_read(tac_fd);
	if(msg != NULL) syslog(LOG_ERR,"TACACS+ stop accounting failed: %s\n", msg);

	close(tac_fd);

	tac_free_attrib(&attr);
	}
   }

int
tacacs_check(void)
    {
    int tac_fd;

    if(!use_tacacs) return -1;

    tac_server[0] = getservername(tac_sname);
    tac_servers++;

    tac_fd=tac_connect(&tac_server[0], tac_servers);
    if(tac_fd < 0) return -1;

    close(tac_fd);
    return 1;
    }
    
void
plugin_init(void)
    {
    char *ptr;
    char *msg;
    struct wordlist *x1;
    struct wordlist *x2;
    
    syslog(LOG_INFO,"TACACS+ 0.5 Init functions");

    tac_server[0] = 0;
    tac_server[1] = 0;
    tac_server[2] = 0;
    tac_sname[0] = '\0';
    tac_coding[0] = '\0';
    tac_secret = tac_coding;
    ptr = ttyname(0);
    if(strncmp(ptr, "/dev/", 5) == 0) ptr += 5;
    strncpy(tty,ptr,32);
    magic_init();
    task_id = (short int)magic();
    add_options(tacacs_options);
    pap_check_hook = tacacs_check;
    pap_auth_hook = tacacs_auth;
    ip_up_hook = accounting_start;
    ip_down_hook = accounting_stop;
    }
